#include <gdk/gdk.h>
#include <math.h>
#include "cairo-image.h"

enum
{
    NO_PROP,
    IMG_FILE,
    IMG_ALPHA,
    IMG_PIXBUF,
    IMG_FIT,
    N_PROPERTIES
};
static GParamSpec *g_prop[N_PROPERTIES] = {0};

struct _DiyaCairoImage
{
    DiyaCairoWidget parent_object;
    gchar * file;
    GdkPixbuf * buffer;
    gdouble alpha;

    gboolean fit;
};

G_DEFINE_FINAL_TYPE(DiyaCairoImage, diya_cairo_image, DIYA_TYPE_CAIRO_WIDGET)

static void diya_cairo_image_dispose(GObject * object)
{
    g_debug("diya_cairo_image_dispose: %s", diya_object_to_string(object));
    DiyaCairoImage * self = DIYA_CAIRO_IMAGE(object);
    if(self->file)
    {
        g_free(self->file);
    }
    if(self->buffer)
    {
        g_debug("diya_cairo_image_dispose: release image buffer");
        g_object_unref(self->buffer);
    }
    G_OBJECT_CLASS(diya_cairo_image_parent_class)->dispose(object);
}

static void diya_cairo_set_pixbuf(DiyaCairoImage* self, GdkPixbuf * buf)
{
    if(self->buffer)
    {
        g_object_unref(self->buffer);
        self->buffer = NULL;
    }
    self->buffer = buf;
    diya_cairo_widget_set_size(self,
        gdk_pixbuf_get_width(self->buffer),
        gdk_pixbuf_get_height(self->buffer)
    );
    diya_cairo_widget_damage(self);
}
static void diya_cairo_load_image_from_file(DiyaCairoImage* self, const gchar* path)
{
    if(self->file)
    {
        g_free(self->file);
    }
    self->file = g_strdup(path);
    GError *error = NULL;
    GdkPixbuf * buf = gdk_pixbuf_new_from_file(self->file,&error);
    if(!buf)
    {
        g_error("diya_cairo_load_image_from_file: [%s] %s", self->file, error->message);
        return;
    }
    diya_cairo_set_pixbuf(self,buf);
}

static void diya_cairo_image_set_property(GObject *object, guint property_id, const GValue *value, GParamSpec *pspec)
{
    DiyaCairoImage * self = DIYA_CAIRO_IMAGE(object);
    switch (property_id)
    {
    case IMG_FILE:
        diya_cairo_load_image_from_file(self, g_value_get_pointer(value));
        break;
    case IMG_ALPHA:
        self->alpha = g_value_get_double(value);
        break;
    case IMG_FIT:
        self->fit = g_value_get_boolean(value);
        break;
    case IMG_PIXBUF:
        if(self->file)
        {
            g_free(self->file);
            self->file = NULL;
        }
        diya_cairo_set_pixbuf(self,g_value_get_pointer(value));
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void diya_cairo_image_get_property(GObject *object, guint property_id, GValue *value, GParamSpec *pspec)
{
    DiyaCairoImage * self = DIYA_CAIRO_IMAGE(object);
    switch (property_id)
    {
    case IMG_FILE:
        g_value_set_pointer(value, self->file);
        break;
    case IMG_ALPHA:
        g_value_set_double(value, self->alpha);
        break;
    case IMG_FIT:
        g_value_set_boolean(value, self->fit);
        break;
    case IMG_PIXBUF:
        g_value_set_pointer(value, self->buffer);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void diya_cairo_image_init(DiyaCairoImage * self)
{
    self->file = NULL;
    self->buffer = NULL;
    self->alpha = 1.0;

    self->fit = FALSE;
}

static gboolean diya_cairo_image_draw(gpointer object, cairo_t * context)
{
    DiyaCairoImage * self = DIYA_CAIRO_IMAGE(object);
    if(!self->buffer || !diya_cairo_widget_is_damaged(self) || !diya_cairo_widget_is_visible(self))
    {
        return FALSE;
    }
    diya_rect_t view_port = diya_cairo_widget_get_view_port(object);
    diya_rect_dump("view port: ", &view_port);
    
    if(diya_rect_is_empty(&view_port))
    {
        return FALSE;
    }
    cairo_save(context);
    cairo_rectangle(context, view_port.x, view_port.y, view_port.width, view_port.height);
    cairo_clip(context);

    diya_color_t * color;
    diya_stroke_t * stroke;
    diya_point_t gpos = diya_cairo_widget_get_global_position(object);
    diya_rect_t * default_extent;
    guint halign, valign;
    diya_offset_t * margin;
    diya_rect_t extent = diya_cairo_widget_get_extent(object);
    diya_rect_dump("extent: ", &extent);
    g_object_get(object,
        "fill", &color,
        "stroke", &stroke,
        "halign", &halign,
        "valign", &valign,
        "margin", &margin,
        "default-extent", &default_extent,
        NULL);
    
    gint img_width = default_extent->width;
    gint img_height = default_extent->height;
    gint stroke_size = ceil(stroke->size);
    GdkPixbuf * buf = NULL;
    if(self->fit)
    {
        img_width = extent.width;
        img_height = extent.height;
        buf = gdk_pixbuf_scale_simple(self->buffer,img_width, img_height, GDK_INTERP_BILINEAR);
    }
    else
    {
        buf = g_object_ref(self->buffer);
    }
    gint off_x = margin->left + stroke->size / 2.0;
    gint off_y = margin->top + stroke->size / 2.0;

    switch (halign)
    {
    case DIYA_CAIRO_WIDGET_ALIGN_MIDDLE:
        off_x += (extent.width - img_width) / 2;
        break;
    case DIYA_CAIRO_WIDGET_ALIGN_END:
        off_x += extent.width - img_width;
        break;
    default:
        break;
    }

    switch (valign)
    {
    case DIYA_CAIRO_WIDGET_ALIGN_MIDDLE:
        off_y += (extent.height - img_height) / 2;
        break;
    case DIYA_CAIRO_WIDGET_ALIGN_END:
        off_y += extent.height - img_height;
        break;
    default:
        break;
    }

    //g_warning("Global position %d %d offset %d %d", gpos.x, gpos.y, off_x, off_y);

    if(color->a > 0)
    {
        //cairo_move_to(context, (double)gpos.x, (double)gpos.y);
        cairo_rectangle(context, 
            gpos.x + margin->left + stroke_size,
            gpos.y + margin->top + stroke_size,
            extent.width - margin->left - margin->right - 2*stroke_size,
            extent.height - margin->bottom - margin->top - 2*stroke_size);
        cairo_set_source_rgba(context, color->r, color->g, color->b, color->a);
        cairo_fill(context);
    }
    // draw image
    gdk_cairo_set_source_pixbuf(context, buf, (double)(gpos.x + off_x), (double)(gpos.y + off_y));
    cairo_paint_with_alpha(context, self->alpha);

    if(stroke->size > 0 && stroke->color.a != 0)
    {
        //cairo_move_to(context, (double)gpos.x, (double)gpos.y);
        cairo_rectangle(context, 
            gpos.x + margin->left + stroke_size,
            gpos.y + margin->top + stroke_size,
            extent.width - margin->left - margin->right - 2*stroke_size,
            extent.height - margin->bottom - margin->top - 2*stroke_size);
        cairo_set_source_rgba(context, stroke->color.r, stroke->color.g, stroke->color.b, stroke->color.a);
        cairo_set_line_width(context,stroke->size);
        cairo_stroke(context);
    }

    g_object_unref(buf);
    cairo_restore(context);
    return TRUE;
}

static void diya_cairo_image_class_init(DiyaCairoImageClass* class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    DiyaCairoWidgetClass *wclass = DIYA_CAIRO_WIDGET_CLASS(class);
    
    wclass->draw = diya_cairo_image_draw;
    
    DIYA_CAIRO_WIDGET_SET_NAME(class, "DiyaCairoImage");
    gobject_class->dispose = diya_cairo_image_dispose;

    gobject_class->set_property = diya_cairo_image_set_property;
    gobject_class->get_property = diya_cairo_image_get_property;
    g_prop[IMG_FILE] = g_param_spec_pointer("file", NULL, "Image file", G_PARAM_READWRITE );
    g_prop[IMG_ALPHA] = g_param_spec_pointer("alpha", NULL, "Image alpha", G_PARAM_READWRITE );
    g_prop[IMG_PIXBUF] = g_param_spec_pointer("pixbuf", NULL, "Image data from gdk_pixbuf", G_PARAM_READWRITE );
    g_prop[IMG_FIT] = g_param_spec_boolean("scale-to-fit", NULL, "Scale image to fit space", FALSE, G_PARAM_READWRITE );
    g_object_class_install_properties (gobject_class, N_PROPERTIES, g_prop);
}